/***********************************************************************

Copyright (c) 2001-2004 Fritz Ganter <ganter@ganter.at>
Website: www.gpsdrive.de

Modifications for ipaq specific version
Copyright (c) 2007,2008 Andrew M. Bishop <amb@gedanken.demon.co.uk>
Website: http://www.gedanken.demon.co.uk/gpsdrive-ipaq/

Disclaimer: Please do not use for navigation. 

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    **********************************************************************/

#include <sys/types.h>
#include <string.h>
#include "gpsdrive.h"
#include <math.h>

/*  Defines for gettext I18n */
# include <libintl.h>
# define _(String) gettext(String)
# ifdef gettext_noop
#  define N_(String) gettext_noop(String)
# else
#  define N_(String) (String)
# endif



/* variables */
extern gint ignorechecksum, debug;
extern gdouble Ra[201];
extern gdouble zero_long, zero_lat;
extern gint SCREEN_X_2, SCREEN_Y_2;
extern gdouble direction;
extern gint xoff, yoff;
extern gdouble current_long, current_lat, groundspeed, milesconv;
extern glong mapscale;

gint
checksum (gchar * text)
{
  gchar t[120], t2[10];
  gint i = 1, checksum = 0, j, orig;

  if (ignorechecksum)
    return TRUE;

  strncpy (t, text, 100);
  t[100] = 0;
  j = strlen (t) - 3;
  while (('\0' != t[i]) && (i < j))
    checksum = checksum ^ t[i++];
  g_strlcpy (t2, (t + j + 1),sizeof(t2));
  sscanf (t2, "%X", &orig);
  if (debug)
    g_print ("\n%s\norigchecksum: %X,my:%X", t, orig, checksum);

  if (orig == checksum)
    {
      g_strlcpy (text, t,1000);
      return TRUE;
    }
  else
    {
     if (debug)
      g_print
	("\n*** NMEA checksum error!\nNMEA: %s\n is: %X, should be: %X\n", t,
	 orig, checksum);
      return FALSE;
    }
}

/* calculates lat and lon for the given position on the screen */
void
calcxytopos (gint posx, gint posy, gdouble * mylat, gdouble * mylon, gint zoom)
{
  gdouble posx_d, posy_d;
  gdouble lat, lon;

  posx_d = (gdouble)(posx - (SCREEN_X_2 - xoff));
  posy_d = (gdouble)((SCREEN_Y_2 - yoff) - posy);

  lon = zero_long + ( posx_d / ( 256 * zoom * (double)(1<<(mapscale-1)) )) * 180;

  lat = 90 * ( atan( tan(M_PI_4*(1+zero_lat/90.0)) * pow(M_E, posy_d * M_PI / ( 256 * zoom * (double)(1<<(mapscale-1)) )))/ M_PI_4 - 1);

  *mylat = lat;
  *mylon = lon;
}

void
calcxy (gint * posx, gint * posy, gdouble lon, gdouble lat, gint zoom)
{
  gdouble posx_d,posy_d;

//  x1 = 256 * zoom * (double)(1<<(mapscale-1)) * (zero_long/180 + 1);
//  x2 = 256 * zoom * (double)(1<<(mapscale-1)) * (lon/180       + 1);

  posx_d = 256 * zoom * (double)(1<<(mapscale-1)) * (lon - zero_long) / 180.0;

//  y1 = 256 * zoom * (double)(1<<(mapscale-1)) * (1 - log(tan(M_PI_4*(1+zero_lat/90.0))) / M_PI );
//  y2 = 256 * zoom * (double)(1<<(mapscale-1)) * (1 - log(tan(M_PI_4*(1+lat/90.0))) / M_PI );

  posy_d = 256 * zoom * (double)(1<<(mapscale-1)) * log(tan(M_PI_4*(1+lat/90.0)) / tan(M_PI_4*(1+zero_lat/90.0))) / M_PI;

//  g_print("xoff,yoff=%d,%d zero = %.4f %.4f : lat,lon %.4f,%.4f -> %.1f,%.1f\n",xoff,yoff,zero_lat,zero_long,lat,lon,posx_d,posy_d);

  posx_d = (SCREEN_X_2 - xoff) + posx_d;
  posy_d = (SCREEN_Y_2 - yoff) - posy_d;

  *posx = (gint)floor(posx_d+0.5);
  *posy = (gint)floor(posy_d+0.5);
}

gdouble
calcR (gdouble lat)
{
  gdouble a = 6378.137, r, sc, x, y, z;
  gdouble e2 = 0.081082 * 0.081082;
/*
the radius of curvature of an ellipsoidal Earth in the plane of the 
meridian is given by 

 R' = a * (1 - e^2) / (1 - e^2 * (sin(lat))^2)^(3/2) 

 where a is the equatorial radius, 
b is the polar radius, and 
e is the eccentricity of the ellipsoid = sqrt(1 - b^2/a^2) 

a = 6378 km (3963 mi) Equatorial radius (surface to center distance) 
b = 6356.752 km (3950 mi) Polar radius (surface to center distance) 
e = 0.081082 Eccentricity 
*/

  lat = lat * M_PI / 180.0;
  sc = sin (lat);
  x = a * (1.0 - e2);
  z = 1.0 - e2 * sc * sc;
  y = pow (z, 1.5);
  r = x / y;

  r = r * 1000.0;
/*      g_print("\nR=%f",r); */
  return r;
}


gdouble
calcdist (gdouble long1, gdouble lat1,gdouble long2, gdouble lat2)
{
  double a, a1, a2, c, d, dlon, dlat, sa, radiant = M_PI / 180;

  dlon = radiant * (long2 - long1);
  dlat = radiant * (lat2 - lat1);

  if ((dlon == 0.0) && (dlat == 0.0))
    return 0.0;

  a1 = sin (dlat / 2);
  a2 = sin (dlon / 2);
  a =
    (a1 * a1) + cos (lat1 * radiant) * cos (lat2 * radiant) * a2 * a2;
  sa = sqrt (a);
  if (sa <= 1.0)
    c = 2 * asin (sa);
  else
    c = 2 * asin (1.0);
  d = (Ra[(int) (100 + lat2)] + Ra[(int) (100 + lat1)]) * c / 2.0;
  return milesconv * d / 1000.0;
}
